@@ -14,8 +14,11 @@
 #include <stdio.h>
 #include <string.h>
 #include <ctype.h>
-#include <wchar.h>
+
+#include <math.h>
 
 #include <NASPRO/core/lib.h>
 
+#include "localepvt.h"
+
 #endif
