@@ -7,8 +7,6 @@
  * See the COPYING file for license conditions.
  */
 
-#define _XOPEN_SOURCE 700
-
 #include "internal.h"
 
 const char *
@@ -49,7 +47,6 @@ nacore_env_path_prefixes_split(const char *path)
 		    != 0)
 		  {
 			nacore_list_free(ret, NULL, NULL);
-			errno = ENOMEM;
 			return NULL;
 		  }
 
@@ -73,7 +70,6 @@ nacore_env_path_prefixes_split(const char *path)
 		    != 0)
 		  {
 			nacore_list_free(ret, NULL, NULL);
-			errno = ENOMEM;
 			return NULL;
 		  }
 	  }
