@@ -164,101 +164,131 @@ nacore_char_utf16le_decode(const char *utf16lec, uint32_t *cp)
 	return 2;
 }
 
-_NACORE_DEF char *
-nacore_string_utf8_to_utf16le(const char *str_utf8)
+_NACORE_DEF size_t
+nacore_string_utf8_to_utf16le_len(const char *str_utf8)
 {
 	const char *c8;
-	char *c16;
-	char *ret;
 	uint32_t cp;
-	size_t size, s;
+	size_t s, ret;
 
-	/* count needed bytes */
-	for (size = 0, c8 = str_utf8; *c8 != '\0'; )
+	for (c8 = str_utf8, ret = 0; *c8 != '\0'; c8 += s)
 	  {
 		s = nacore_char_utf8_decode(c8, &cp);
 		if (s == 0)
 		  {
-			c8++;
+			s = 1;
 			continue;
 		  }
 
-		size += nacore_char_utf16le_encode(NULL, cp);
-		c8 += s;
+		ret += nacore_char_utf16le_encode(NULL, cp);
 	  }
 
-	/* allocate */
-	ret = malloc(size + 2);
-	if (ret == NULL)
-		return NULL;
-
-	/* convert */
-	for (c8 = str_utf8, c16 = ret; *c8 != '\0'; )
-	  {
-		s = nacore_char_utf8_decode(c8, &cp);
-		if (s == 0)
-		  {
-			c8++;
-			continue;
-		  }
-
-		size = nacore_char_utf16le_encode(c16, cp);
-		c8 += s;
-		c16 += size;
-	  }
-
-	/* null termination */
-	c16[0] = 0;
-	c16[1] = 0;
-
 	return ret;
 }
 
-_NACORE_DEF char *
-nacore_string_utf16le_to_utf8(const char *str_utf16le)
+_NACORE_DEF void
+nacore_string_utf8_to_utf16le(char *buf, const char *str_utf8)
 {
-	const char *c16;
-	char *c8;
-	char *ret;
+	const char *c8;
+	char *c16;
 	uint32_t cp;
-	size_t size, s;
+	size_t s8, s16;
 
-	/* count needed bytes */
-	for (size = 0, c16 = str_utf16le; *c16 != '\0'; )
+	for (c8 = str_utf8, c16 = buf; *c8 != '\0'; c8 += s8, c16 += s16)
 	  {
-		s = nacore_char_utf16le_decode(c16, &cp);
-		if (s == 0)
+		s8 = nacore_char_utf8_decode(c8, &cp);
+		if (s8 == 0)
 		  {
-			c16 += 2;
+			s8 = 1;
+			s16 = 0;
 			continue;
 		  }
 
-		size += nacore_char_utf8_encode(NULL, cp);
-		c16 += s;
+		s16 = nacore_char_utf16le_encode(c16, cp);
 	  }
 
-	/* allocate */
-	ret = malloc(size + 1);
+	c16[0] = 0;
+	c16[1] = 0;
+}
+
+_NACORE_DEF char *
+nacore_string_utf8_to_utf16le_a(const char *str_utf8)
+{
+	char *ret;
+	size_t len;
+
+	len = nacore_string_utf8_to_utf16le_len(str_utf8);
+
+	ret = malloc(len + 2);
 	if (ret == NULL)
 		return NULL;
 
-	/* convert */
-	for (c16 = str_utf16le, c8 = ret; *c16 != '\0'; )
+	nacore_string_utf8_to_utf16le(ret, str_utf8);
+
+	return ret;
+}
+
+_NACORE_DEF size_t
+nacore_string_utf16le_to_utf8_len(const char *str_utf16le)
+{
+	const char *c16;
+	uint32_t cp;
+	size_t s, ret;
+
+	for (c16 = str_utf16le, ret = 0; (c16[0] != 0) || (c16[1] != 0);
+	     c16 += s)
 	  {
 		s = nacore_char_utf16le_decode(c16, &cp);
 		if (s == 0)
 		  {
-			c16 += 2;
+			s = 2;
 			continue;
 		  }
 
-		size = nacore_char_utf8_encode(c8, cp);
-		c16 += s;
-		c8 += size;
+		ret += nacore_char_utf8_encode(NULL, cp);
+	  }
+
+	return ret;
+}
+
+_NACORE_DEF void
+nacore_string_utf16le_to_utf8(char *buf, const char *str_utf16le)
+{
+	const char *c16;
+	char *c8;
+	uint32_t cp;
+	size_t s8, s16;
+
+	for (c8 = buf, c16 = str_utf16le; (c16[0] != 0) || (c16[1] != 0);
+	     c8 += s8, c16 += s16)
+	  {
+		s16 = nacore_char_utf16le_decode(c16, &cp);
+		if (s16 == 0)
+		  {
+			s16 = 2;
+			s8 = 0;
+			continue;
+		  }
+
+		s8 = nacore_char_utf8_encode(c8, cp);
 	  }
 
-	/* null termination */
 	*c8 = '\0';
+}
+
+_NACORE_DEF char *
+nacore_string_utf16le_to_utf8_a(const char *str_utf16le)
+{
+	char *ret;
+	size_t len;
+
+	len = nacore_string_utf16le_to_utf8_len(str_utf16le);
+
+	ret = malloc(len + 1);
+	if (ret == NULL)
+		return NULL;
+
+	nacore_string_utf16le_to_utf8(ret, str_utf16le);
 
 	return ret;
 }
@@ -281,390 +311,8 @@ nacore_strnlen(const char *s, size_t maxlen)
 	return ret;
 }
 
-/* Code point category type for grapheme breaking. */
-typedef enum
-  {
-	cp_other = 0,
-	cp_cr,
-	cp_lf,
-	cp_control,
-	cp_extend,
-	cp_prepend,
-	cp_spacing_mark,
-	cp_l,
-	cp_v,
-	cp_t,
-	cp_lv,
-	cp_lvt
-  } code_point_category;
-
-/* Rules for breaking graphemes. Each index of the array represents the current
- * code point category, each bit position inside an array element represents the
- * next code point category, both according to the code_point_category
- * enumeration. When the bit is set to 1 there is no grapheme break. */
-static uint16_t
-grapheme_breaking_chart[12] =
-  {0x50, 0x04, 0, 0, 0x50, 0xff1, 0x50, 0xdd0, 0x350, 0x250, 0x350, 0x250};
-
-/* Gets the category for grapheme breaking of the code point pointed by s.
- * 
- * TODO: needs a reimplementation using some FSM (finite state machine)
- * representation. */
-static code_point_category
-get_code_point_category(const char *s)
-{
-	uint32_t v;
-
-	if (*s == 0xd)
-		return cp_cr;
-	else if (*s == 0xa)
-		return cp_lf;
-	else if ((*s & 0x80) == 0)	/* one byte */
-	  {
-		if (((*s >= 0x0) && (*s <= 0x9)) || (*s == 0xb) || (*s == 0xc)
-		    || ((*s >= 0xe) && (*s <= 0x1f)) || (*s == 0x7f))
-			return cp_control;
-	  }
-	else if ((*s & 0xe0) == 0xc0)	/* two bytes */
-	  {
-		v = ((s[0] & 0x1f) << 6) | (s[1] & 0x3f);
-
-		if (((v >= 0x600) && (v <= 0x603))
-		    || (v == 0x6dd) || (v == 0x70f))
-			return cp_control;
-
-		if (((v >= 0x300) && (v <= 0x36f))
-		    || ((v >= 0x483) && (v <= 0x489))
-		    || ((v >= 0x591) && (v <= 0x5bd))
-		    || ((v >= 0x5bf) && (v <= 0x5c2))
-		    || (v == 0x5c4) || (v == 0x5c5) || (v == 0x5c7)
-		    || ((v >= 0x610) && (v <= 0x61a))
-		    || ((v >= 0x64b) && (v <= 0x65f)) || (v == 0x670)
-		    || ((v >= 0x646) && (v <= 0x6dc))
-		    || ((v >= 0x6df) && (v <= 0x6e4))
-		    || (v == 0x6e7) || (v == 0x6e8)
-		    || ((v >= 0x6ea) && (v <= 0x6ed)) || (v == 0x711)
-		    || ((v >= 0x730) && (v <= 0x74a))
-		    || ((v >= 0x7a6) && (v <= 0x7b0))
-		    || ((v >= 0x7eb) && (v <= 0x7f3)))
-			return cp_extend;
-	  }
-	else if ((*s & 0xf0) == 0xe0)	/* three bytes */
-	  {
-		v = ((s[0] & 0xf) << 12) | ((s[1] & 0x3f) << 6) | (s[2] & 0x3f);
-
-		if ((v == 0x17b4) || (v == 0x17b5) || (v == 0x200b)
-		    || (v == 0x200e) || (v == 0x200f) || (v == 0x2028)
-		    || ((v >= 0x2029) && (v <= 0x202e))
-		    || ((v >= 0x2060) && (v <= 0x2064))
-		    || ((v >= 0x206a) && (v <= 0x206f))
-		    || (v == 0xfeff) || ((v >= 0xfff9) && (v <= 0xfffb)))
-			return cp_control;
-
-		if (((v >= 0x816) && (v <= 0x82d)  && (v != 0x820)
-		     && (v != 0x824) && (v != 0x828))
-		    || ((v >= 0x859) && (v <= 0x85b))
-		    || ((v >= 0x900) && (v <= 0x902))
-		    || (v == 0x93a) || (v == 0x93c)
-		    || ((v >= 0x941) && (v <= 0x948)) || (v == 0x94d)
-		    || ((v >= 0x951) && (v <= 0x957))
-		    || (v == 0x962) || (v == 0x963) || (v == 0x981)
-		    || (v == 0x9bc) || (v == 0x9bc) || (v == 0x9be)
-		    || ((v >= 0x9c1) && (v <= 0x9c4))
-		    || (v == 0x9cd) || (v == 0x9d7) || (v == 0x9e2)
-		    || (v == 0x9e3) || (v == 0xa01) || (v == 0xa02)
-		    || (v == 0xa3c) || (v == 0xa41) || (v == 0xa42)
-		    || (v == 0xa47) || (v == 0xa48)
-		    || ((v >= 0xa4b) && (v <= 0xa4d)) || (v == 0xa51)
-		    || (v == 0xa70) || (v == 0xa71) || (v == 0xa75)
-		    || (v == 0xa81) || (v == 0xa82) || (v == 0xabc)
-		    || ((v >= 0xac1) && (v <= 0xac8) && (v != 0xac6))
-		    || (v == 0xacd) || (v == 0xae2) || (v == 0xae3)
-		    || (v == 0xb01) || (v == 0xb3c) || (v == 0xb3e)
-		    || (v == 0xb3f) || ((v >= 0xb41) && (v <= 0xb44))
-		    || (v == 0xb4d) || (v == 0xb56) || (v == 0xb57)
-		    || (v == 0xb62) || (v == 0xb63) || (v == 0xb82)
-		    || (v == 0xbbe) || (v == 0xbc0) || (v == 0xbcd)
-		    || (v == 0xbd7) || ((v >= 0xc3e) && (v <= 0xc40))
-		    || ((v >= 0xc46) && (v <= 0xc48))
-		    || ((v >= 0xc4a) && (v <= 0xc4d))
-		    || (v == 0xc55) || (v == 0xc56) || (v == 0xc62)
-		    || (v == 0xc63) || (v == 0xcbc) || (v == 0xcbf)
-		    || (v == 0xcc2) || (v == 0xcc6) || (v == 0xccc)
-		    || (v == 0xccd) || (v == 0xcd5) || (v == 0xcd6)
-		    || (v == 0xce2) || (v == 0xce3) || (v == 0xd3e)
-		    || ((v >= 0xd41) && (v <= 0xd44)) || (v == 0xd4d)
-		    || (v == 0xd57) || (v == 0xd62) || (v == 0xd63)
-		    || (v == 0xdca) || (v == 0xdcf)
-		    || ((v >= 0xdd2) && (v <= 0xdd4)) || (v == 0xdd6)
-		    || (v == 0xddf) || (v == 0xe31)
-		    || ((v >= 0xe34) && (v <= 0xe3a))
-		    || ((v >= 0xe47) && (v <= 0xe4e)) || (v == 0xeb1)
-		    || ((v >= 0xeb4) && (v <= 0xeb9))
-		    || (v == 0xebb) || (v == 0xebc)
-		    || ((v >= 0xec8) && (v <= 0xecd)) || (v == 0xf18)
-		    || (v == 0xf19) || (v == 0xf35) || (v == 0xf37)
-		    || (v == 0xf39)
-		    || ((v >= 0xf71) && (v <= 0xf87) && (v != 0xf7f)
-			&& (v != 0xf85))
-		    || ((v >= 0xf8d) && (v <= 0xfbc) && (v != 0xf98))
-		    || (v == 0xfc6)
-		    || ((v >= 0x102d) && (v <= 0x103a) && (v != 0x1031)
-			&& (v != 0x1038)) || (v == 0x103d) || (v == 0x103e)
-		    || (v == 0x1058) || (v == 0x1059)
-		    || ((v >= 0x105e) && (v <= 0x1060))
-		    || ((v >= 0x1071) && (v <= 0x1074)) || (v == 0x1082)
-		    || (v == 0x1085) || (v == 0x1086) || (v == 0x108d)
-		    || (v == 0x109d) || ((v >= 0x135d) && (v <= 0x135f))
-		    || ((v >= 0x1712) && (v <= 0x1714))
-		    || ((v >= 0x1732) && (v <= 0x1734)) || (v == 0x1752)
-		    || (v == 0x1753) || (v == 0x1772) || (v == 0x1773)
-		    || ((v >= 0x17b7) && (v <= 0x17bd)) || (v == 0x17c6)
-		    || ((v >= 0x17c9) && (v <= 0x17d3)) || (v == 0x17dd)
-		    || ((v >= 0x180b) && (v <= 0x180d)) || (v == 0x18a9)
-		    || ((v >= 0x1920) && (v <= 0x1922)) || (v == 0x1927)
-		    || (v == 0x1928) || (v == 0x1932)
-		    || ((v >= 0x1939) && (v <= 0x193b)) || (v == 0x1a17)
-		    || (v == 0x1a18) || (v == 0x1a56)
-		    || ((v >= 0x1a58) && (v <= 0x1a5e)) || (v == 0x1a60)
-		    || (v == 0x1a62) || ((v >= 0x1a65) && (v <= 0x1a6c))
-		    || ((v >= 0x1a73) && (v <= 0x1a7c)) || (v == 0x1a7f)
-		    || ((v >= 0x1b00) && (v <= 0x1b03))
-		    || ((v >= 0x1b34) && (v <= 0x1b3c) && (v != 0x1b35)
-			&& (v != 0x1b3b)) || (v == 0x1b42)
-		    || ((v >= 0x1b6b) && (v <= 0x1b73)) || (v == 0x1b80)
-		    || (v == 0x1b81) || ((v >= 0x1ba2) && (v <= 0x1ba5))
-		    || (v == 0x1ba8) || (v == 0x1ba9) || (v == 0x1be6)
-		    || (v == 0x1be8) || (v == 0x1be9)
-		    || ((v >= 0x1bed) && (v <= 0x1bf1) && (v != 0x1bee))
-		    || ((v >= 0x1c2c) && (v <= 0x1c33))
-		    || (v == 0x1c36) || (v == 0x1c37)
-		    || ((v >= 0x1cd0) && (v <= 0x1ce8) && (v != 0x1cd3)
-			&& (v != 0x1ce1)) || (v == 0x1ced)
-		    || ((v >= 0x1dc0) && (v <= 0x1de6))
-		    || ((v >= 0x1dfc) && (v <= 0x1dff))
-		    || (v == 0x200c) || (v == 0x200d)
-		    || ((v >= 0x20d0) && (v <= 0x20f0))
-		    || ((v >= 0x2cef) && (v <= 0x2cf1))
-		    || (v == 0x2d7f) || ((v >= 0x2de0) && (v <= 0x2dff))
-		    || ((v >= 0x302a) && (v <= 0x302f))
-		    || ((v >= 0x3099) && (v <= 0x309a))
-		    || ((v >= 0xa66f) && (v <= 0xa672)) || (v == 0xa67c)
-		    || (v == 0xa67d) || (v == 0xa6f0) || (v == 0xa6f1)
-		    || (v == 0xa802) || (v == 0xa806) || (v == 0xa80b)
-		    || (v == 0xa825) || (v == 0xa826) || (v == 0xa8c4)
-		    || ((v >= 0xa8e0) && (v <= 0xa8f1))
-		    || ((v >= 0xa926) && (v <= 0xa92d))
-		    || ((v >= 0xa947) && (v <= 0xa951))
-		    || ((v >= 0xa980) && (v <= 0xa982)) || (v == 0xa9b3)
-		    || ((v >= 0xa9b6) && (v <= 0xa9b9)) || (v == 0xa9bc)
-		    || ((v >= 0xaa29) && (v <= 0xaa2e)) || (v == 0xaa31)
-		    || (v == 0xaa32) || (v == 0xaa35) || (v == 0xaa36)
-		    || (v == 0xaa45) || (v == 0xaa4c) || (v == 0xaab0)
-		    || ((v >= 0xaab2) && (v <= 0xaab4)) || (v == 0xaab7)
-		    || (v == 0xaab8) || (v == 0xaabe) || (v == 0xaabf)
-		    || (v == 0xaac1) || (v == 0xabe5) || (v == 0xabe8)
-		    || (v == 0xabed) || (v == 0xfb1e)
-		    || ((v >= 0xfe00) && (v <= 0xfe0f))
-		    || ((v >= 0xfe20) && (v <= 0xfe26))
-		    || (v == 0xff9e) || (v == 0xff9f))
-			return cp_extend;
-
-		if (((v >= 0xe40) && (v <= 0xe44))
-		    || ((v >= 0xec0) && (v <= 0xec4))
-		    || (v == 0xaab5) || (v == 0xaab6)
-		    || ((v >= 0xaab9) && (v <= 0xaabc) && (v != 0xaaba)))
-			return cp_prepend;
-
-		if ((v == 0x903) || (v == 0x93b)
-		    || ((v >= 0x93e) && (v <= 0x940))
-		    || ((v >= 0x949) && (v <= 0x94c)) || (v == 0x94e)
-		    || (v == 0x94f) || (v == 0x982) || (v == 0x983)
-		    || ((v >= 0x9bf) && (v <= 0x9c8)) || (v == 0x9cb)
-		    || (v == 0x9cc) || (v == 0xa03)
-		    || ((v >= 0xa3e) && (v <= 0xa40)) || (v == 0xa83)
-		    || ((v >= 0xabe) && (v <= 0xac0))
-		    || ((v >= 0xac9) && (v <= 0xacc) && (v != 0xaca))
-		    || (v == 0xb02) || (v == 0xb03) || (v == 0xb40)
-		    || (v == 0xb47) || (v == 0xb48) || (v == 0xb4b)
-		    || (v == 0xb4c) || (v == 0xbbf) || (v == 0xbc1)
-		    || (v == 0xbc2) || ((v >= 0xbc6) && (v <= 0xbc8))
-		    || ((v >= 0xbca) && (v <= 0xbcc))
-		    || ((v >= 0xc01) && (v <= 0xc03))
-		    || ((v >= 0xc41) && (v <= 0xc44)) || (v == 0xc82)
-		    || (v == 0xc83) || (v == 0xcbe) || (v == 0xcc0)
-		    || (v == 0xcc1) || (v == 0xcc3) || (v == 0xcc4)
-		    || (v == 0xcc7) || (v == 0xcc8) || (v == 0xcca)
-		    || (v == 0xccb) || (v == 0xd02) || (v == 0xd03)
-		    || (v == 0xd3f) || (v == 0xd40)
-		    || ((v >= 0xd46) && (v <= 0xd4c) && (v != 0xd49))
-		    || (v == 0xd82) || (v == 0xd83) || (v == 0xdd0)
-		    || (v == 0xdd1) || ((v >= 0xdd8) && (v <= 0xdde))
-		    || (v == 0xdf2) || (v == 0xdf3) || (v == 0xe30)
-		    || (v == 0xe32) || (v == 0xe33) || (v == 0xe45)
-		    || (v == 0xeb0) || (v == 0xeb2) || (v == 0xeb3)
-		    || (v == 0xf3e) || (v == 0xf3f) || (v == 0xf7f)
-		    || (v == 0x120b) || (v == 0x120c) || (v == 0x1031)
-		    || (v == 0x1038) || (v == 0x103b) || (v == 0x103c)
-		    || (v == 0x1056) || (v == 0x1057)
-		    || ((v >= 0x1062) && (v <= 0x1064))
-		    || ((v >= 0x1067) && (v <= 0x106d))
-		    || (v == 0x1083) || (v == 0x1084)
-		    || ((v >= 0x1087) && (v <= 0x108c))
-		    || ((v >= 0x108f) && (v <= 0x109c)) || (v == 0x17b6)
-		    || ((v >= 0x17be) && (v <= 0x17c8) && (v != 0x17c6))
-		    || ((v >= 0x1923) && (v <= 0x1926))
-		    || ((v >= 0x1929) && (v <= 0x192b)) || (v == 0x1930)
-		    || (v == 0x1931) || ((v >= 0x1933) && (v <= 0x1938))
-		    || ((v >= 0x19b0) && (v <= 0x19c0)) || (v == 0x19c8)
-		    || (v == 0x19c9) || ((v >= 0x1a19) && (v <= 0x1a1b))
-		    || (v == 0x1a55) || (v == 0x1a57) || (v == 0x1a61)
-		    || (v == 0x1a63) || (v == 0x1a64)
-		    || ((v >= 0x1a6d) && (v <= 0x1a72))
-		    || (v == 0x1b04) || (v == 0x1b35)
-		    || ((v >= 0x1b3b) && (v <= 0x1b44) && (v != 0x1b3c)
-			&& (v != 0x1b42))
-		    || (v == 0x1b82) || (v == 0x1ba1) || (v == 0x1ba6)
-		    || (v == 0x1ba7) || (v == 0x1baa) || (v == 0x1be7)
-		    || ((v >= 0x1bea) && (v <= 0x1bec)) || (v == 0x1bee)
-		    || (v == 0x1bf2) || (v == 0x1bf3)
-		    || ((v >= 0x1c24) && (v <= 0x1c2b)) || (v == 0x1c34)
-		    || (v == 0x1c35) || (v == 0x1ce1) || (v == 0x1cf2)
-		    || (v == 0xa823) || (v == 0xa824) || (v == 0xa827)
-		    || (v == 0xa880) || (v == 0xa881)
-		    || ((v >= 0xa8b4) && (v <= 0xa8c3)) || (v == 0xa952)
-		    || (v == 0xa953) || (v == 0xa983) || (v == 0xa9b4)
-		    || (v == 0xa9b5) || (v == 0xa9ba) || (v == 0xa9bb)
-		    || ((v >= 0xa9bd) && (v <= 0xa9c0)) || (v == 0xaa2f)
-		    || (v == 0xaa30) || (v == 0xaa33) || (v == 0xaa34)
-		    || (v == 0xaa4d) || (v == 0xaa7b)
-		    || ((v >= 0xabe3) && (v <= 0xabec) && (v != 0xabe5)
-			&& (v != 0xabe8) && (v != 0xabeb)))
-			return cp_spacing_mark;
-
-		if (((v >= 0x1100) && (v <= 0x115f))
-		    || ((v >= 0xa960) && (v <= 0xa97c)))
-			return cp_l;
-
-		if (((v >= 0x1160) && (v <= 0x11a7))
-		    || ((v >= 0xd7b0) && (v <= 0xd7c6)))
-			return cp_v;
-
-		if (((v >= 0x11a8) && (v <= 0x11ff))
-		    || ((v >= 0xd7cb) && (v <= 0xd7fb)))
-			return cp_t;
-
-		if ((v >= 0xac00) && (v <= 0xd788) && ((v % 0x1c) == 0x10))
-			return cp_lv;
-
-		if ((v >= 0xac01) && (v <= 0xd7a3))
-			return cp_lvt;
-	  }
-	else				/* four bytes */
-	  {
-		v = ((s[0] & 0x7) << 18) | ((s[1] & 0x3f) << 12)
-		    | ((s[2] & 0x3f) << 6) | (s[3] & 0x3f);
-
-		if ((v == 0x110bd) || (v == 0x1d173) || (v == 0x1d174)
-		    || (v == 0xe0001) || ((v >= 0xe0020) && (v <= 0xe007f)))
-			return cp_control;
-
-		if ((v == 0x101fd)
-		    || ((v >= 0x10a01) && (v <= 0x10a06) && (v != 0x10a04))
-		    || ((v >= 0x10a0c) && (v <= 0x10a0f))
-		    || ((v >= 0x10a38) && (v <= 0x10a3a))
-		    || (v == 0x10a3f) || (v == 0x11001)
-		    || ((v >= 0x11038) && (v <= 0x11046))
-		    || (v == 0x11080) || (v == 0x11081)
-		    || ((v >= 0x110b3) && (v <= 0x110b6))
-		    || ((v >= 0x110b9) && (v <= 0x110ba))
-		    || ((v >= 0x1d165) && (v <= 0x1d169) && (v != 0x1d166))
-		    || ((v >= 0x1d16e) && (v <= 0x1d172))
-		    || ((v >= 0x1d17b) && (v <= 0x1d182))
-		    || ((v >= 0x1d185) && (v <= 0x1d18b))
-		    || ((v >= 0x1d1aa) && (v <= 0x1d1ad))
-		    || ((v >= 0x1d242) && (v <= 0x1d244))
-		    || ((v >= 0xe0100) && (v <= 0x1e01ef)))
-			return cp_extend;
-
-		if ((v == 0x11000) || (v == 0x11002) || (v == 0x11082)
-		    || ((v >= 0x110b0) && (v <= 0x110b2)) || (v == 0x110b7)
-		    || (v == 0x110b8) || (v == 0x1d166) || (v == 0x1d16d))
-			return cp_spacing_mark;
-	  }
-
-	return cp_other;
-}
-
-_NACORE_DEF size_t
-nacore_strgraphemes(const char *s)
-{
-	size_t ret;
-	const char *c, *n;
-	code_point_category cpc, cpn;
-
-	if (*s == '\0')
-		return 0;
-
-	ret = 0;
-	n = s;
-	c = NULL;
-	cpc = 0;
-	while (*n != '\0')
-	  {
-		cpn = get_code_point_category(n);
-
-		if (c != NULL)
-			if (!(grapheme_breaking_chart[cpc] & (1 << cpn)))
-				ret++;
-
-		c = n;
-		cpc = cpn;
-
-		n += (*((unsigned char *)n) < 0x80)
-		     ? 1 : ((*((unsigned char *)n) < 0xe0)
-			   ? 2 : ((*((unsigned char *)n) < 0xf0) ? 3 : 4));
-	  }
-
-	return ret + 1;
-}
-
-_NACORE_DEF size_t
-nacore_strngraphemes(const char *s, size_t max)
-{
-	size_t ret;
-	const char *c, *n;
-	code_point_category cpc, cpn;
-
-	if ((*s == '\0') || (max == 0))
-		return 0;
-
-	ret = 0;
-	n = s;
-	c = NULL;
-	cpc = 0;
-	max--;
-	while ((*n != '\0') && (ret < max))
-	  {
-		cpn = get_code_point_category(n);
-
-		if (c != NULL)
-			if (!(grapheme_breaking_chart[cpc] & (1 << cpn)))
-				ret++;
-
-		c = n;
-		cpc = cpn;
-
-		n += (*((unsigned char *)n) < 0x80)
-		     ? 1 : ((*((unsigned char *)n) < 0xe0)
-			   ? 2 : ((*((unsigned char *)n) < 0xf0) ? 3 : 4));
-	  }
-
-	return ret + 1;
-}
-
 _NACORE_DEF char *
-nacore_astrcpy(const char *s, void *unused)
+nacore_strdup(const char *s, void *unused)
 {
 	char *ret;
 	size_t len;
@@ -1106,7 +754,8 @@ pad_right(char **str, conv_spec *cs, int len)
 /* Conversion functions. */
 
 static int
-conv_d(char *str, conv_spec *cs, va_list *ap)
+conv_d(char *str, conv_spec *cs, va_list *ap, const char *dec_sep,
+       size_t dec_sep_len)
 {
 	intmax_t v, c;
 	int ret, i;
@@ -1198,7 +847,8 @@ conv_d(char *str, conv_spec *cs, va_list *ap)
 }
 
 static int
-conv_o(char *str, conv_spec *cs, va_list *ap)
+conv_o(char *str, conv_spec *cs, va_list *ap, const char *dec_sep,
+       size_t dec_sep_len)
 {
 	uintmax_t v, c;
 	int ret, i;
@@ -1273,7 +923,8 @@ conv_o(char *str, conv_spec *cs, va_list *ap)
 }
 
 static int
-conv_u(char *str, conv_spec *cs, va_list *ap)
+conv_u(char *str, conv_spec *cs, va_list *ap, const char *dec_sep,
+       size_t dec_sep_len)
 {
 	uintmax_t v, c;
 	int ret, i;
@@ -1421,19 +1072,22 @@ conv_xX(char *str, conv_spec *cs, va_list *ap, char X)
 }
 
 static int
-conv_x(char *str, conv_spec *cs, va_list *ap)
+conv_x(char *str, conv_spec *cs, va_list *ap, const char *dec_sep,
+       size_t dec_sep_len)
 {
 	return conv_xX(str, cs, ap, 0);
 }
 
 static int
-conv_X(char *str, conv_spec *cs, va_list *ap)
+conv_X(char *str, conv_spec *cs, va_list *ap, const char *dec_sep,
+       size_t dec_sep_len)
 {
 	return conv_xX(str, cs, ap, 1);
 }
 
 static int
-conv_eE(char *str, conv_spec *cs, va_list *ap, char E)
+conv_eE(char *str, conv_spec *cs, va_list *ap, const char *dec_sep,
+	size_t dec_sep_len, char E)
 {
 	long double v, p, d;
 	short int exp, dexp, pe, de;
@@ -1444,12 +1098,11 @@ conv_eE(char *str, conv_spec *cs, va_list *ap, char E)
 	if (cs->precision == -1)
 		cs->precision = 6;
 
+	ret = 3;
 	exp = 0; 
 	dexp = 2;
 
-	if (isinf(v) || isnan(v))
-		ret = 3;
-	else
+	if (!isinf(v) && !isnan(v))
 	  {
 		if (v < -0.0)
 		  {
@@ -1466,10 +1119,8 @@ conv_eE(char *str, conv_spec *cs, va_list *ap, char E)
 			exp = floor(log10l(v));
 		  }
 
-		if ((cs->precision == 0) && !(cs->flags & FLAG_DASH))
-			ret = 3;
-		else
-			ret = 4 + cs->precision;
+		if ((cs->precision != 0) || (cs->flags & FLAG_DASH))
+			ret += dec_sep_len + cs->precision;
 
 		if (exp < 0)
 			dexp = floor(log10(-exp) + 1.0);
@@ -1528,8 +1179,8 @@ conv_eE(char *str, conv_spec *cs, va_list *ap, char E)
 
 		if ((cs->precision != 0) || (cs->flags & FLAG_DASH))
 		  {
-			*str = '.';
-			str++;
+			strncpy(str, dec_sep, dec_sep_len);
+			str += dec_sep_len;
 
 			for (; cs->precision != 0; cs->precision--)
 			  {
@@ -1555,8 +1206,8 @@ conv_eE(char *str, conv_spec *cs, va_list *ap, char E)
 
 		if ((cs->precision != 0) || (cs->flags & FLAG_DASH))
 		  {
-			*str = '.';
-			str++;
+			strncpy(str, dec_sep, dec_sep_len);
+			str += dec_sep_len;
 
 			for (; cs->precision != 0; cs->precision--)
 			  {
@@ -1597,19 +1248,22 @@ conv_eE(char *str, conv_spec *cs, va_list *ap, char E)
 }
 
 static int
-conv_e(char *str, conv_spec *cs, va_list *ap)
+conv_e(char *str, conv_spec *cs, va_list *ap, const char *dec_sep,
+       size_t dec_sep_len)
 {
-	return conv_eE(str, cs, ap, 0);
+	return conv_eE(str, cs, ap, dec_sep, dec_sep_len, 0);
 }
 
 static int
-conv_E(char *str, conv_spec *cs, va_list *ap)
+conv_E(char *str, conv_spec *cs, va_list *ap, const char *dec_sep,
+       size_t dec_sep_len)
 {
-	return conv_eE(str, cs, ap, 1);
+	return conv_eE(str, cs, ap, dec_sep, dec_sep_len, 1);
 }
 
 static int
-conv_fF(char *str, conv_spec *cs, va_list *ap, char F)
+conv_fF(char *str, conv_spec *cs, va_list *ap, const char *dec_sep,
+	size_t dec_sep_len, char F)
 {
 	long double v, p;
 	short int d;
@@ -1647,7 +1301,7 @@ conv_fF(char *str, conv_spec *cs, va_list *ap, char F)
 		ret += d;
 
 		if ((cs->precision != 0) || (cs->flags & FLAG_DASH))
-			ret += 1 + cs->precision;
+			ret += dec_sep_len + cs->precision;
 	  }
 
 	if (signbit(v) || (cs->flags & FLAG_PLUS) || (cs->flags & FLAG_SPACE))
@@ -1712,8 +1366,8 @@ conv_fF(char *str, conv_spec *cs, va_list *ap, char F)
 
 		if ((cs->precision != 0) || (cs->flags & FLAG_DASH))
 		  {
-			*str = '.';
-			str++;
+			strncpy(str, dec_sep, dec_sep_len);
+			str += dec_sep_len;
 
 			for (; cs->precision != 0; cs->precision--)
 			  {
@@ -1732,19 +1386,22 @@ conv_fF(char *str, conv_spec *cs, va_list *ap, char F)
 }
 
 static int
-conv_f(char *str, conv_spec *cs, va_list *ap)
+conv_f(char *str, conv_spec *cs, va_list *ap, const char *dec_sep,
+       size_t dec_sep_len)
 {
-	return conv_fF(str, cs, ap, 0);
+	return conv_fF(str, cs, ap, dec_sep, dec_sep_len, 0);
 }
 
 static int
-conv_F(char *str, conv_spec *cs, va_list *ap)
+conv_F(char *str, conv_spec *cs, va_list *ap, const char *dec_sep,
+       size_t dec_sep_len)
 {
-	return conv_fF(str, cs, ap, 1);
+	return conv_fF(str, cs, ap, dec_sep, dec_sep_len, 1);
 }
 
 static int
-conv_gG(char *str, conv_spec *cs, va_list *ap, char G)
+conv_gG(char *str, conv_spec *cs, va_list *ap, const char *dec_sep,
+	size_t dec_sep_len, char G)
 {
 	long double v, vc, p, d, min;
 	short int exp, dexp, pe, de;
@@ -1757,12 +1414,11 @@ conv_gG(char *str, conv_spec *cs, va_list *ap, char G)
 	else if (cs->precision == 0)
 		cs->precision = 1;
 
+	ret = 3;
 	exp = 0;
 	dexp = 2;
 
-	if (isinf(v) || isnan(v))
-		ret = 3;
-	else
+	if (!isinf(v) && !isnan(v))
 	  {
 		vc = v; /* make gcc happy */
 
@@ -1786,11 +1442,9 @@ conv_gG(char *str, conv_spec *cs, va_list *ap, char G)
 			v = vc;
 
 			if (cs->flags & FLAG_DASH)
-				ret = 4 + cs->precision;
+				ret += dec_sep_len + cs->precision;
 			else
 			  {
-				ret = 3;
-
 				p = powl(10.0, exp);
 				vc = v;
 				d = vc / p;
@@ -1838,7 +1492,7 @@ conv_gG(char *str, conv_spec *cs, va_list *ap, char G)
 				ret = 1;
 
 			if (cs->flags & FLAG_DASH)
-				ret += 1 + cs->precision;
+				ret += dec_sep_len + cs->precision;
 			else
 			  {
 				vc = fabsl(vc);
@@ -1927,8 +1581,8 @@ conv_gG(char *str, conv_spec *cs, va_list *ap, char G)
 			if ((cs->flags & FLAG_DASH)
 			    || ((v >= min) || (v <= -min)))
 			  {
-				*str = '.';
-				str++;
+				strncpy(str, dec_sep, dec_sep_len);
+				str += dec_sep_len;
 
 				for (; (cs->precision != 0)
 				       && ((cs->flags & FLAG_DASH)
@@ -1995,8 +1649,8 @@ conv_gG(char *str, conv_spec *cs, va_list *ap, char G)
 			if ((cs->flags & FLAG_DASH)
 			    || ((v <= -min) || (v >= min)))
 			  {
-				*str = '.';
-				str++;
+				strncpy(str, dec_sep, dec_sep_len);
+				str += dec_sep_len;
 
 				p = 0.1L;
 				for (; (cs->precision != 0)
@@ -2022,19 +1676,22 @@ conv_gG(char *str, conv_spec *cs, va_list *ap, char G)
 }
 
 static int
-conv_g(char *str, conv_spec *cs, va_list *ap)
+conv_g(char *str, conv_spec *cs, va_list *ap, const char *dec_sep,
+       size_t dec_sep_len)
 {
-	return conv_gG(str, cs, ap, 0);
+	return conv_gG(str, cs, ap, dec_sep, dec_sep_len, 0);
 }
 
 static int
-conv_G(char *str, conv_spec *cs, va_list *ap)
+conv_G(char *str, conv_spec *cs, va_list *ap, const char *dec_sep,
+       size_t dec_sep_len)
 {
-	return conv_gG(str, cs, ap, 1);
+	return conv_gG(str, cs, ap, dec_sep, dec_sep_len, 1);
 }
 
 static int
-conv_aA(char *str, conv_spec *cs, va_list *ap, char A)
+conv_aA(char *str, conv_spec *cs, va_list *ap, const char *dec_sep,
+	size_t dec_sep_len, char A)
 {
 	long double v, p, d, vc;
 	short int exp, dexp, pe, de;
@@ -2071,10 +1728,9 @@ conv_aA(char *str, conv_spec *cs, va_list *ap, char A)
 				exp = floor(log2l(v));
 			  }
 
-			if ((cs->precision == 0) && !(cs->flags & FLAG_DASH))
-				ret = 5;
-			else
-				ret = 6 + cs->precision;
+			ret = 5;
+			if ((cs->precision != 0) || (cs->flags & FLAG_DASH))
+				ret += dec_sep_len + cs->precision;
 		  }
 		else
 		  {
@@ -2088,7 +1744,7 @@ conv_aA(char *str, conv_spec *cs, va_list *ap, char A)
 
 			if (((vc != 0.0) && (vc != -0.0))
 			    || (cs->flags & FLAG_DASH))
-				ret++;
+				ret += dec_sep_len;
 
 			while ((vc != 0.0) && (vc != -0.0))
 			  {
@@ -2169,8 +1825,8 @@ conv_aA(char *str, conv_spec *cs, va_list *ap, char A)
 		if ((cs->precision != -1)
 		    && ((cs->precision != 0) || (cs->flags & FLAG_DASH)))
 		  {
-			*str = '.';
-			str++;
+			strncpy(str, dec_sep, dec_sep_len);
+			str += dec_sep_len;
 
 			for (; cs->precision != 0; cs->precision--)
 			  {
@@ -2185,8 +1841,8 @@ conv_aA(char *str, conv_spec *cs, va_list *ap, char A)
 		  }
 		else if ((v != 0.0) && (v != -0.0) && (cs->precision == -1))
 		  {
-			*str = '.';
-			str++;
+			strncpy(str, dec_sep, dec_sep_len);
+			str += dec_sep_len;
 
 			while ((v != 0.0) && (v != -0.0))
 			  {
@@ -2229,20 +1885,23 @@ conv_aA(char *str, conv_spec *cs, va_list *ap, char A)
 }
 
 static int
-conv_a(char *str, conv_spec *cs, va_list *ap)
+conv_a(char *str, conv_spec *cs, va_list *ap, const char *dec_sep,
+       size_t dec_sep_len)
 {
-	return conv_aA(str, cs, ap, 0);
+	return conv_aA(str, cs, ap, dec_sep, dec_sep_len, 0);
 }
 
 static int
-conv_A(char *str, conv_spec *cs, va_list *ap)
+conv_A(char *str, conv_spec *cs, va_list *ap, const char *dec_sep,
+       size_t dec_sep_len)
 {
-	return conv_aA(str, cs, ap, 1);
+	return conv_aA(str, cs, ap, dec_sep, dec_sep_len, 1);
 }
 
 /* FIXME: wint_t ("%lc") not supported */
 static int
-conv_c(char *str, conv_spec *cs, va_list *ap)
+conv_c(char *str, conv_spec *cs, va_list *ap, const char *dec_sep,
+       size_t dec_sep_len)
 {
 	unsigned char v;
 
@@ -2263,7 +1922,8 @@ conv_c(char *str, conv_spec *cs, va_list *ap)
 
 /* FIXME: const wchar_t * ("%ls") not supported */
 static int
-conv_s(char *str, conv_spec *cs, va_list *ap)
+conv_s(char *str, conv_spec *cs, va_list *ap, const char *dec_sep,
+       size_t dec_sep_len)
 {
 	const char *v;
 	int ret;
@@ -2287,24 +1947,27 @@ conv_s(char *str, conv_spec *cs, va_list *ap)
 }
 
 static int
-conv_p(char *str, conv_spec *cs, va_list *ap)
+conv_p(char *str, conv_spec *cs, va_list *ap, const char *dec_sep,
+       size_t dec_sep_len)
 {
 	cs->flags |= FLAG_DASH;
 	cs->lm = lm_l;
 
-	return conv_x(str, cs, ap);
+	return conv_x(str, cs, ap, dec_sep, dec_sep_len);
 }
 
 /* Conversion functions' table (index of function according to
  * enum conv_specifier. */
-static int (*conv_func[])(char *str, conv_spec *cs, va_list *ap) =
+static int (*conv_func[])(char *str, conv_spec *cs, va_list *ap,
+			  const char *dec_sep, size_t dec_sep_len) =
   {
 	conv_d, conv_o, conv_u, conv_x, conv_X, conv_e, conv_E, conv_f, conv_F,
 	conv_g, conv_G, conv_a, conv_A, conv_c, conv_s, conv_p
   };
 
-_NACORE_DEF NACORE_FORMAT_VPRINTF(2) int
-nacore_vasprintf(char **strp, const char *fmt, va_list ap)
+static int
+_vasprintf(char **strp, const char *fmt, va_list ap, char *dec_sep,
+	   size_t dec_sep_len)
 {
 	int ret;
 	const char *c;
@@ -2332,7 +1995,8 @@ nacore_vasprintf(char **strp, const char *fmt, va_list ap)
 		c++;
 		cs = get_conv_spec(&c, &ap2);
 		if (cs.cs != cs_n)
-			ret += conv_func[cs.cs](NULL, &cs, &ap2);
+			ret += conv_func[cs.cs](NULL, &cs, &ap2, dec_sep,
+						dec_sep_len);
 	  }
 
 	va_end(ap2);
@@ -2374,7 +2038,8 @@ nacore_vasprintf(char **strp, const char *fmt, va_list ap)
 		  }
 		else
 		  {
-			i = conv_func[cs.cs](s, &cs, &ap2);
+			i = conv_func[cs.cs](s, &cs, &ap2, dec_sep,
+					     dec_sep_len);
 			ret += i;
 			s += i;
 		  }
@@ -2388,14 +2053,48 @@ nacore_vasprintf(char **strp, const char *fmt, va_list ap)
 	return ret;
 }
 
+_NACORE_DEF NACORE_FORMAT_VPRINTF(2) int
+nacore_vasprintf(char **strp, const char *fmt, va_list ap)
+{
+	size_t dec_sep_len = _nacore_locale_dec_sep_len();
+	char dec_sep[dec_sep_len + 1];
+
+	_nacore_locale_dec_sep_get(dec_sep);
+
+	return _vasprintf(strp, fmt, ap, dec_sep, dec_sep_len);
+}
+
+_NACORE_DEF NACORE_FORMAT_VPRINTF(2) int
+nacore_vasprintf_nl(char **strp, const char *fmt, va_list ap)
+{
+	return _vasprintf(strp, fmt, ap, ".", 1);
+}
+
 _NACORE_DEF NACORE_FORMAT_PRINTF(2, 3) int
 nacore_asprintf(char **strp, const char *fmt, ...)
 {
 	va_list ap;
 	int ret;
+	size_t dec_sep_len = _nacore_locale_dec_sep_len();
+	char dec_sep[dec_sep_len + 1];
+
+	_nacore_locale_dec_sep_get(dec_sep);
 
 	va_start(ap, fmt);
-	ret = nacore_vasprintf(strp, fmt, ap);
+	ret = _vasprintf(strp, fmt, ap, dec_sep, dec_sep_len);
+	va_end(ap);
+
+	return ret;
+}
+
+_NACORE_DEF NACORE_FORMAT_PRINTF(2, 3) int
+nacore_asprintf_nl(char **strp, const char *fmt, ...)
+{
+	va_list ap;
+	int ret;
+
+	va_start(ap, fmt);
+	ret = _vasprintf(strp, fmt, ap, ".", 1);
 	va_end(ap);
 
 	return ret;
@@ -2459,6 +2158,5 @@ err:
 	if (buf != NULL)
 		free(buf);
 	nacore_list_free(ret, NULL, NULL);
-	errno = ENOMEM;
 	return NULL;
 }
