@@ -7,6 +7,9 @@
  * See the COPYING file for license conditions.
  */
 
+#define _WIN32_WINNT	0x501
+#define UNICODE
+
 #include "internal.h"
 
 #include <windows.h>
@@ -19,7 +22,7 @@ nacore_env_get(const char *name)
 	char *namew;
 	DWORD err;
 
-	namew = nacore_string_utf8_to_utf16le(name);
+	namew = nacore_string_utf8_to_utf16le_a(name);
 	if (namew == NULL)
 	  {
 		errno = ENOMEM;
@@ -35,7 +38,7 @@ nacore_env_get(const char *name)
 		return NULL;
 	  }
 
-	ret = nacore_string_utf16le_to_utf8(buf);
+	ret = nacore_string_utf16le_to_utf8_a(buf);
 	if (ret == NULL)
 	  {
 		errno = ENOMEM;
